"""Helper gates for circuit knitting."""

from qiskit import QuantumCircuit, transpile
from qiskit.circuit import CircuitInstruction, Gate, Instruction

# define the cut location marker
cut_op = QuantumCircuit(1, name="Cut")
cut_op = cut_op.to_instruction(label="Cut")
cut_op.definition = None

def cut() -> QuantumCircuit |Instruction:
    """Return a single qubit wire cut instruction."""
    return cut_op


cutCZ_op = QuantumCircuit(2, name="CutCZ")
cutCZ_op = cutCZ_op.to_instruction(label="CutCZ")
cutCZ_op.definition = None

def cutCZ() -> QuantumCircuit | Instruction:
    """Return a two qubit cutCZ gate instruction."""
    return cutCZ_op


def cutGate(gate: Gate, control: int, target: int) -> dict:
    """Return a cutCZ circuit with the same parameters as the input gate."""
    if gate.num_qubits != 2:
        raise ValueError("Input gate must be a 2-qubit gate.")
    if control == target:
        raise ValueError("Control and target qubits must be different.")
    if control < 0 or target < 0:
        raise ValueError("Control and target qubits must be non-negative.")

    qc = QuantumCircuit(2, name=f"cut{gate.name.upper()}")
    if control > target:
        loccontrol = 1
        loctarget = 0
    else:
        loccontrol = 0
        loctarget = 1
    qc.append(gate, [loccontrol, loctarget])
    tr = transpile(qc, basis_gates=["cz", "r", "h", "s", "sdg", "x", "y", "z"])
    for ind, instr in enumerate(tr.data):
        if instr.operation.name == "cz":
            tr.data.pop(ind)
            test = CircuitInstruction(
                        operation=cutCZ_op,
                        qubits=tr.qubits,
                    )
            tr.data.insert(ind, test)
    return {"instruction": tr.to_instruction(label="CutGate"), 
            "qargs":[control, target]}

# define measurements for different bases
xmeas = QuantumCircuit(1, 1, name="x-meas")
xmeas.h(0)
xmeas.measure(0, 0)

ymeas = QuantumCircuit(1, 1, name="y-meas")
ymeas.sdg(0)
ymeas.h(0)
ymeas.measure(0, 0)

idmeas = QuantumCircuit(1, name="id-meas")
idmeas.id(0)

zmeas = QuantumCircuit(1, 1, name="z-meas")
zmeas.measure(0, 0)


# define initialization operations
zero_init = QuantumCircuit(1, name="0-init")
zero_init.id(0)

one_init = QuantumCircuit(1, name="1-init")
one_init.x(0)

plus_init = QuantumCircuit(1, name="'+'-init")
plus_init.h(0)

minus_init = QuantumCircuit(1, name="'-'-init")
minus_init.h(0)
minus_init.z(0)

i_plus_init = QuantumCircuit(1, name="'i+'-init")
i_plus_init.h(0)
i_plus_init.s(0)

i_minus_init = QuantumCircuit(1, name="'i-'-init")
i_minus_init.h(0)
i_minus_init.z(0)
i_minus_init.s(0)

sdg = QuantumCircuit(1, 1, name="sdg")
sdg.sdg(0)

s = QuantumCircuit(1, 1, name="s")
s.s(0)

z = QuantumCircuit(1, 1, name="z")
z.z(0)

sdg_meas = QuantumCircuit(1, 1, name="sdg_meas")
sdg_meas.sdg(0)
sdg_meas.measure(0, 0)


"""Config."""  # noqa: INP001
# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import os
import sys

# Find the path to the source files we want to document, relative to the location of this file,
# convert it to an absolute path.
py_path = os.path.join(os.getcwd(), os.path.dirname(__file__), '../')
sys.path.insert(0, os.path.abspath(py_path))

project = "QCut"
copyright = "2024, Joonas Nivala"
author = "Joonas Nivala"

# The short X.Y version.
version = ''
# The full version, including alpha/beta/rc tags.
release = ''
try:
    from QCut import __version__ as version
except ImportError:
    pass
else:
    release = version

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
    "sphinx_rtd_theme",
    "myst_nb",
    "sphinx.ext.mathjax",
    "sphinx.ext.githubpages",
]

autodoc_default_options = {
    'members': True,
    'member-order': 'bysource',
    'special-members': '__init__',
    'undoc-members': False,
    'show-inheritance': True,
    'private-members': False,
}

autosummary_generate = True
autosummary_ignore_patterns = []

templates_path = ["_templates"]
exclude_patterns = [
    "_build",
    "Thumbs.db",
    ".DS_Store",
    "jupyter_execute",
    "build",
    "test",
    ".venv",
    "docs-venv",
]

source_suffix = {
    '.rst': 'restructuredtext',
    '.ipynb': 'myst-nb',
    '.md': 'myst-nb',
}

myst_enable_extensions = [
    "amsmath",
    "dollarmath",
]

nb_execution_mode = "off"

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_book_theme"
html_logo = "_static/images/QCut-logo.jpg"
html_title = "QCut"
html_favicon = "_static/images/QCut-logo.jpg"
html_static_path = ["_static"]



import qiskit
from qiskit import QuantumCircuit
from qiskit.circuit.library import CXGate
from qiskit.quantum_info import SparsePauliOp

import QCut as ck
from QCut import cut, cutGate

cut_circ = QuantumCircuit(4)

mult = 1.635
cut_circ.r(mult*0.46262, mult*0.1446, 0)
cut_circ.append(**cutGate(CXGate(), 0, 1)) 
cut_circ.append(cut(), [1])
cut_circ.cx(1,2)
cut_circ.cx(2,3)

if qiskit.__version__ < "2.2.0":
        from iqm.qiskit_iqm import IQMFakeAdonis
        backend = IQMFakeAdonis()

else:
    from qiskit.providers.fake_provider import GenericBackendV2
    backend = GenericBackendV2(num_qubits=5, basis_gates=["cz", "r"])

def test_transpile_subcircuits():
    """Test transpilation of subcircuits with custom gates.

    This test verifies that the transpile_subcircuits function correctly transpiles
    subcircuits containing custom gates (cutCZ and cutGate) to a specified backend.
    It checks that the transpiled circuits contain the expected gates and that the
    custom gates are preserved in the transpilation process.
    """

    cut_copy = cut_circ.copy()

    cut_circuit = ck.get_locations_and_subcircuits(cut_copy)

    transpiled = ck.transpile_subcircuits(cut_circuit, backend, optimization_level=3)

    assert transpiled.num_subcircuits == 3

    assert transpiled.backend == backend

    basis = ["cz", "r"]

    for subcircuit in transpiled.subcircuits:
        for op in subcircuit.data:
            assert (op.operation.name.lower() in basis 
                    or "cut" in op.operation.name.lower() 
                    or "meas" in op.operation.name.lower()
                    or "init" in op.operation.name.lower()
                    or "obs" in op.operation.name.lower())
            
def test_transpile_experiments():
    """Test transpilation of experiments with custom gates.

    This test verifies that the transpile_experiments function correctly transpiles
    experiments containing custom gates (cutCZ and cutGate) to a specified backend.
    It checks that the transpiled circuits contain the expected gates and that the
    custom gates are preserved in the transpilation process.
    """

    cut_copy = cut_circ.copy()

    cut_circuit = ck.get_locations_and_subcircuits(cut_copy)

    exp_circuits = ck.get_experiment_circuits(cut_circuit, SparsePauliOp("ZZZZ"))

    transpiled = ck.transpile_experiments(exp_circuits, backend, optimization_level=3)

    assert transpiled.backend == backend

    assert transpiled.num_circuits == exp_circuits.num_circuits

    basis = ["cz", "r"]

    for exp in transpiled.experiments:
        for exp_dict in exp:
            for circ in exp_dict.values():
                for op in circ.data:
                    assert (op.operation.name.lower() in basis 
                            or "cut" in op.operation.name.lower() 
                            or "meas" in op.operation.name.lower()
                            or "init" in op.operation.name.lower()
                            or "obs" in op.operation.name.lower())


"""Init circuit knitting."""  # noqa: N999

from importlib.metadata import PackageNotFoundError, version

from QCut.backend_utility import (
    transpile_experiments,
    transpile_subcircuits,
)
from QCut.circuit_knitting import (
    get_experiment_circuits,
    run,
    run_cut_circuit,
    run_experiments,
)
from QCut.circuit_preparation import (
    get_locations_and_subcircuits,
)
from QCut.postprocess import (
    estimate_expectation_values,
)
from QCut.QCutFind import find_cuts
from QCut.qpd_gates import cut, cutCZ, cutGate

try:
    __version__ = version("QCut")
except PackageNotFoundError:  # pragma: no cover
    __version__ = "unknown"
finally:
    del version, PackageNotFoundError

__all__ = [
    "run_on_backend",
    "transpile_experiments",
    "transpile_subcircuits",
    "estimate_expectation_values",
    "get_experiment_circuits",
    "get_locations_and_subcircuits",
    "run",
    "run_cut_circuit",
    "run_experiments",
    "cut",
    "cutCZ",
    "cutGate",
    "find_cuts"
]



"""Class for nicely representing a cut circuit/experiment. Also implements 
some of the same functionality as the qiskit QuantumCircuit class for
a group of circuts."""

from __future__ import annotations

from qiskit import QuantumCircuit

from QCut.cutlocation import CutLocation, SingleQubitCutLocation
    

class CutCircuit:
    """Class for representing a cut circuit. Contains the subcircuits, cut locations,
    and mapping of qubits. Also contains some of the same functionality as the qiskit 
    QuantumCircuit class for a group of circuits."""

    def __init__(self, 
                    subcircuits: list[QuantumCircuit],
                    cut_locations: list[CutLocation | SingleQubitCutLocation],
                    map_qubit: dict[int, int],
                    backend=None) -> None:
        """Init."""

        self.subcircuits = subcircuits
        self.cut_locations = cut_locations
        self.map_qubit = map_qubit
        self.backend = backend

    def assign_parameters(self, parameters: dict, inplace=False
                          ) -> CutCircuit | None:
        """Assign parameters to the circuits. Same as qiskit
        QuantumCircuit.assign_parameters."""
        if inplace:
            for ind, circuit in enumerate(self.subcircuits):
                try:
                    self.subcircuits[ind] = circuit.assign_parameters(parameters)
                except Exception:
                    pass
            return 
        
        else:
            new_circuits = []
            for ind, circuit in enumerate(self.subcircuits):
                try:
                    new_circuits.append(circuit.assign_parameters(parameters))
                except Exception:
                    new_circuits.append(circuit)
            return CutCircuit(new_circuits,
                              self.cut_locations, self.map_qubit, self.backend)
            
    @property
    def num_qubits(self):
        """Number of qubits per subcircuit."""
        return [i.num_qubits for i in self.subcircuits]
    
    @property
    def num_subcircuits(self):
        """Total number of circuits."""
        return len(self.subcircuits)


class CutExperiment:
    def __init__(self, 
                    experiment_circuits: list[list[dict[int, QuantumCircuit]]],
                    cut_locations: list[CutLocation | SingleQubitCutLocation],
                    map_qubit: dict[int, int],
                    coefficients: list[float],
                    observables,
                    backend=None) -> None:
        """Init."""

        self.experiments = experiment_circuits
        self.backend = backend
        self.cut_locations = cut_locations
        self.map_qubit = map_qubit
        self.coefficients = coefficients
        self.observables = observables

    def expv_data(self):
        """Get data for expv calculation."""
        return {
            "cut_locations": self.cut_locations,
            "map_qubit": self.map_qubit,
            "coefficients": self.coefficients,
            "observables": self.observables
        }

    def assign_parameters(self, parameters: dict, inplace=False
                          ) -> CutExperiment | None:
        """Assign parameters to the circuits. Same as qiskit
        QuantumCircuit.assign_parameters."""
        if inplace:
            for exp_ind, subcircuits in enumerate(self.experiments):
                for circ_ind, value in enumerate(subcircuits):
                    for ind, circuit in value.items():
                        try:
                            self.experiments[exp_ind][circ_ind][ind] = (circuit
                                                            .assign_parameters(parameters))
                        except Exception:
                            pass
            return
        else:
            new_experiments = []
            for exp_ind, subcircuits in enumerate(self.experiments):
                new_subcircuits = []
                for circ_ind, value in enumerate(subcircuits):
                    new_circuits = {}
                    for ind, circuit in value.items():
                        try:
                            new_circuits[ind] = (circuit
                                                .assign_parameters(parameters))
                        except Exception:
                            new_circuits[ind] = circuit
                    new_subcircuits.append(new_circuits)
                new_experiments.append(new_subcircuits)
            return CutExperiment(new_experiments,
                                 self.cut_locations, self.map_qubit,
                                 self.coefficients, self.observables, self.backend)
    
    @property
    def num_qubits(self):
        """Number of qubits per subcircuit."""
        return [i.num_qubits for i in self.experiments[0][0].values()]
    
    @property
    def num_circuits(self):
        """Total number of circuits."""
        return sum(len(subcircuits) * len(subcircuits[0]) 
                   for subcircuits in self.experiments)
    
    @property
    def group_size(self):
        """Number of circuits in a group."""
        return len(self.experiments[0][0][0])
    
    @property
    def num_groups(self):
        """Number of circuit groups."""
        return len(self.experiments)
    
    @property
    def num_obs_groups(self):
        return len(self.experiments[0])


""" 
Module for transforming measurement bases in quantum circuits 
when generating experiment circuits, particularly for combining Pauli 
operators into compatible measurement settings.
"""

from __future__ import annotations

from copy import deepcopy

from qiskit import QuantumCircuit
from qiskit.circuit import (
    Instruction,
    QuantumRegister,
)
from qiskit.circuit.library import HGate, SdgGate
from qiskit.dagcircuit import DAGCircuit
from qiskit.quantum_info import SparsePauliOp
from qiskit.transpiler import PassManager
from qiskit.transpiler.basepasses import TransformationPass


def _combine_pauli_ops(op: SparsePauliOp) -> list[dict[int, str]]:  # noqa: C901
    """Combine Pauli operators that have no conflicting non-identity components.
    
    Args:
        op (SparsePauliOp): The SparsePauliOp to analyze.
    
    Returns:
        list[dict[int, str]]: A list of combined measurement settings, where each dict
                              maps qubit indices to Pauli basis measurements.
    """

    pauli_strings = [pauli.to_label()[::-1] for pauli in op.paulis]
    
    combined_settings = []
    used = [False] * len(pauli_strings)
    
    for i, pauli_string in enumerate(pauli_strings):
        if used[i]:
            continue
        
        # Start a new combined setting with the current Pauli string
        combined = {}
        for qubit_index, pauli in enumerate(pauli_string):
            if pauli != "I":
                combined[qubit_index] = pauli
        
        used[i] = True
        
        # Try to combine with remaining Pauli strings
        for j in range(i + 1, len(pauli_strings)):
            if used[j]:
                continue
            
            # Check if pauli_strings[j] can be combined with current combined setting
            can_combine = True
            for qubit_index, pauli in enumerate(pauli_strings[j]):
                if pauli != "I":
                    if qubit_index in combined and combined[qubit_index] != pauli:
                        can_combine = False
                        break
            
            # If compatible, add to combined setting
            if can_combine:
                for qubit_index, pauli in enumerate(pauli_strings[j]):
                    if pauli != "I":
                        combined[qubit_index] = pauli
                used[j] = True
        
        combined_settings.append(combined)
    
    return combined_settings

class _ModifyMeasurementBasis(TransformationPass):
 
    def __init__(
        self,
        measurement_settings: list[dict[int, str]],
        ops: dict[str, Instruction] | None = None,

    ):
        
        self.measurement_settings = measurement_settings
        self.ops = ops
        super().__init__()
 
    def run(
        self,
        dag: DAGCircuit,
    ) -> DAGCircuit:
        
        no_obs = True
        cloned_dag = deepcopy(dag)
        for node in dag.op_nodes():
            if "obs" not in node.op.name:
                continue
 
            obs_ind = int(node.op.name.split("_")[-1])
            
            for setting in self.measurement_settings:
                if obs_ind not in setting:
                    #continue
                    dag.remove_op_node(node)
                    break
                
                ob = setting[obs_ind]
                no_obs = False
                
                mini_dag = DAGCircuit()
                register = QuantumRegister(1)
                mini_dag.add_qreg(register)

                if ob == "X":
                    if self.ops and "X-meas" in self.ops:
                        mini_dag.apply_operation_back(
                            self.ops["X-meas"], [register[0]]
                        )
                    else:
                        mini_dag.apply_operation_back(
                            HGate(), [register[0]]
                        )
                elif ob == "Y":
                    if self.ops and "Y-meas" in self.ops:
                        mini_dag.apply_operation_back(
                            self.ops["Y-meas"], [register[0]]
                        )
                    else:
                        mini_dag.apply_operation_back(
                            SdgGate(), [register[0]]
                        )
                        mini_dag.apply_operation_back(
                            HGate(), [register[0]]
                        )
                
                dag.substitute_node_with_dag(node, mini_dag)

        if no_obs:
            return cloned_dag
        return dag

def _get_obs_subcircuits(subcircuits: list[QuantumCircuit], 
                        measurement_settings: list[dict[int, str]],
                        ops: dict[str, Instruction] | None = None
                        ) -> list[dict[int, QuantumCircuit]]:
    pms = [PassManager([_ModifyMeasurementBasis([setting], ops)]) 
           for setting in measurement_settings]
    obs_subcircuits = []
    for pm in pms:
        pm_circs = {}
        for ind, subcircuit in enumerate(subcircuits):
            modified_circuit = pm.run(subcircuit)
            if modified_circuit.num_qubits == 0:
                continue
            pm_circs[ind] = modified_circuit
        obs_subcircuits.append(pm_circs)
    return obs_subcircuits

def _get_observable_circuit_index(pauli, combined: list[dict[int, str]]):
    """Find which measurement setting covers the non-identity letters of `pauli`,
    and return the indices of the qubits involved."""
    label = pauli
    non_identity = {i: p for i, p in enumerate(label) if p.to_label() != "I"}

    for idx, setting in enumerate(combined):
        # All non-identity qubits must be measured in the matching basis
        if all(setting.get(q) == p.to_label() for q, p in non_identity.items()):
            return {"circuit_index": idx, "obs_indices": list(non_identity.keys())}

    return {"circuit_index": None, "obs_indices": []}


from qiskit import QuantumCircuit
from qiskit.circuit.library import CXGate
from qiskit.quantum_info import SparsePauliOp
from qiskit_aer import AerSimulator

import QCut as ck
from QCut import cut, cutGate

circuit  =  QuantumCircuit(4)

mult = 1.635
circuit.r(mult*0.46262, mult*0.1446, 0)
circuit.cx(0,1)
circuit.cx(1,2)
circuit.cx(2,3)
   

cut_circuit = QuantumCircuit(4)

mult = 1.635
cut_circuit.r(mult*0.46262, mult*0.1446, 0)
cut_circuit.append(**cutGate(CXGate(), 0, 1)) 
cut_circuit.append(cut(), [1])
cut_circuit.cx(1,2)
cut_circuit.cx(2,3)

subcirc_lens = [3,4,6]

res_expvs = [0.727323, 0.727323, 0.727323, 1.000000]

def test_cut_gate_num_subcircuits():
    cut_circ = cut_circuit.copy()

    cut_qc = ck.get_locations_and_subcircuits(cut_circ)

    assert cut_qc.num_subcircuits == 3

def test_cut_gate_subcircuits():
    cut_qc = ck.get_locations_and_subcircuits(cut_circuit.copy())

    for ind, circ in enumerate(cut_qc.subcircuits):
        assert len(circ.data) == subcirc_lens[ind]

def test_cut_gate_expectation_values():
    cut_qc = ck.get_locations_and_subcircuits(cut_circuit.copy())

    backend = AerSimulator()

    observables = SparsePauliOp(["IIIZ", "IIZI", "IZII", "IIZZ"])

    cut_experiment = ck.get_experiment_circuits(cut_qc, observables)

    results = ck.run_experiments(cut_experiment, backend=backend)

    expectation_values = ck.estimate_expectation_values(results, 
                                                        cut_experiment.expv_data())

    for ind, expv in enumerate(expectation_values):
        assert abs(expv - res_expvs[ind]) < 0.1


"""Storage class for storing cut locations."""

from __future__ import annotations

from qiskit import QuantumRegister


class CutLocation:
    """Storage class for storing cut locations."""

    def __init__(
        self, cut_location: tuple[list, int]
    ) -> None:
        """Init."""
        self.qubits = cut_location[0]
        self.control = cut_location[0][0][1]
        self.target = cut_location[0][1][1]
        self.index = cut_location[1]

    def __eq__(self, other) -> bool:
        """Equality."""
        if not isinstance(other, CutLocation):
            return NotImplemented

        return (
            self.control == other.control
            and self.target == other.target
            and self.index == other.index
        )

    def __str__(self) -> str:
        """Format string."""
        msg = (
            f"control qubit: {self.control}, target qubit: {self.target}, "
            f"cut index: {self.index}"
        )
        return msg

    def __repr__(self) -> str:
        """Represent as string."""
        return str(self)

class SingleQubitCutLocation:
    """Storage class for storing cut locations."""

    def __init__(self, cut_location: tuple[tuple[QuantumRegister, int], int]) -> None:
        """Init."""
        self.qubits = cut_location[0]
        self.meas = cut_location[0][1]
        self.init = cut_location[0][1]
        self.index = cut_location[1]

    def __eq__(self, other) -> bool:
        """Equality."""
        if not isinstance(other, SingleQubitCutLocation):
            return NotImplemented

        return (
            self.meas == other.meas
            and self.init == other.init
            and self.index == other.index
        )

    def __str__(self) -> str:
        """Format string."""
        msg = (
            f"meas qubit: {self.meas}, init qubit: {self.init}, "
            f"cut index: {self.index}"
        )
        return msg

    def __repr__(self) -> str:
        """Represent as string."""
        return str(self)    


"""
Utility functions for running on real backends.
"""

from __future__ import annotations

from qiskit import transpile
from qiskit.circuit import Gate
from qiskit.transpiler import Target

from QCut.circuit_utils import _remove_idle_wires
from QCut.cutcircuit import CutCircuit, CutExperiment
from QCut.cutlocation import CutLocation, SingleQubitCutLocation


def transpile_subcircuits(cut_circuit: CutCircuit,
                          backend,
                          optimization_level: int = 0,
                          transpile_options: dict | None = None) -> CutCircuit:
    """
    Transpile subcircuits for a given backend. More efficient than transpiling
    experiment circuits as it only transpiles each subcircuit once instead of
    each experiment circuit. However, may lead to suboptimal transpilation results as
    the tranpiler cannot use the backend object directly due to need to retain
    some placeholder gates for cuts and observables. `transpile_options` can be used
    to pass additional options to the transpiler. For more control over transpilation
    of experiment circuits, use `transpile_experiments` or manually transpile them.

    Args:
        subcircuits (list[QuantumCircuit]): List of subcircuits to be transpiled.
        backend: Backend to transpile to.
        optimization_level (int): Optimization level for transpilation (0-3).
        transpile_options (dict): Arguments passed to qiskit transpile function.
    Returns:
        CutCircuit: Transpiled subcircuits wrapped in CutCircuit class.
    """

    if not isinstance(cut_circuit, CutCircuit):
        raise ValueError("cut_circuit must be of type CutCircuit.")

    custom_gates = {}
    for ind, i in enumerate(cut_circuit.cut_locations):
        if isinstance(i, CutLocation):
            custom_gates[f"cutCZ_t_{ind}"] = Gate(
                num_qubits=1, name=f"cutCZ_t_{ind}", params=[], label=f"cutCZ_t_{ind}"
            )
            custom_gates[f"cutCZ_c_{ind}"] = Gate(
                num_qubits=1, name=f"cutCZ_c_{ind}", params=[], label=f"cutCZ_c_{ind}"
            )

        elif isinstance(i, SingleQubitCutLocation):
            custom_gates[f"Meas_{ind}"] = Gate(
                num_qubits=1, name=f"Meas_{ind}", params=[], label=f"Meas_{ind}"
            )
            custom_gates[f"Init_{ind}"] = Gate(
                num_qubits=1, name=f"Init_{ind}", params=[], label=f"Init_{ind}"
            )
    
    for i in range(sum([x.num_qubits for x in cut_circuit.subcircuits])):
        custom_gates[f"obs_{i}"] = Gate(
            num_qubits=1, name=f"obs_{i}", params=[], label=f"obs_{i}"
        )

    target = Target()

    try:
        basis_gates = list({i[0].name for i in backend._target.instructions})
    except Exception:
        return cut_circuit
    
    target = target.from_configuration(
            num_qubits=backend.num_qubits,
            coupling_map=backend._coupling_map,
            basis_gates=basis_gates + list(custom_gates.keys()),
            custom_name_mapping=custom_gates,
        )


    transpiled = transpile(cut_circuit.subcircuits,
                           target=target,
                           optimization_level=optimization_level,
                           **(transpile_options or {}))

    transpiled = [_remove_idle_wires(circ) for circ in transpiled]

    return CutCircuit(subcircuits=transpiled, cut_locations=cut_circuit.cut_locations, 
                      map_qubit=cut_circuit.map_qubit, backend=backend)

def transpile_experiments(cut_experiment: CutExperiment, 
                          backend,
                          optimization_level: int = 0,
                          transpile_options: dict | None = None) -> CutExperiment:
    """
    Transpile experiment circuits. Transpiles all generated experiment circuits for
    a given backend. Most often one should use `transpile_subcircuits` instead, as that
    only transpiles subcircuits before experiment generation which is a lot more 
    efficient. This function is mainly provided for special cases where one needs/wants
    extra control over the transpilation of experiment circuits.

    Args:
        cut_experiment: (CutExperiment): Experiment circuits to be transpiled.
        backend (str): Backend to transpile to.
        optimization_level (int): Optimization level for transpilation (0-3).
        transpile_options (dict): Arguments passed to qiskit transpile function.

    Returns:
        CutExperiment: Transpiled experiment circuits wrapped in CutExperiment class.
    """

    if not isinstance(cut_experiment, CutExperiment):
        raise ValueError("cut_experiment must be of type CutExperiment.")

    subexperiments = [
        [
            {
                ind: transpile(circ,
                               backend=backend,
                               optimization_level=optimization_level,
                               **(transpile_options or {}))
                for ind, circ in exp.items()
            }
            for exp in exps
        ]
        for exps in cut_experiment.experiments
    ]

    return CutExperiment(subexperiments,
                         cut_locations=cut_experiment.cut_locations,
                         map_qubit=cut_experiment.map_qubit,
                         coefficients=cut_experiment.coefficients,
                         observables=cut_experiment.observables,
                         backend=backend)



"""
A module for post-processing the results obtained from running the cut circuits and
calculating the estimated expectation values based on the results and the 
provided observables.
"""

from __future__ import annotations

from itertools import product
from typing import Optional

import numpy as np

from QCut.basis_transform import _combine_pauli_ops, _get_observable_circuit_index
from QCut.cutlocation import SingleQubitCutLocation
from QCut.qcutresult import SubResult, TotalResult

ERROR = 0.0000001

def _process_results(
    results: list,
    shots: int,
    samples: int,
) -> list[list[TotalResult]]:
    """Transform results with post processing function {0,1} -> [-1, 1].

    Tranform results so that we map 0 -> -1 and 1 -> 1. Gives processed results in form
    [TotalResult0, TotalResult1, ...], where TotalResult is
    [SubResult0, SubResult1, ...] and SubResult are [[[x0,x0,x0], [y0,y0], counts0],
    [[x1,x1,x1], [y1,y1], counts1], ...].

    Args:
        results (list): results from experiment circuits
        shots (int): number of shots per circuit run
        samples (int): number of needed samples

    Returns:
    -------
        list[TotalResult]:
            list of transformed results

    """
    preocessed_results = []

    for group_ind, circ_group in enumerate(results):
        for exp_ind, experiment_run in enumerate(circ_group):
            experiment_run_results = []
            for sub_ind, sub_result in experiment_run.items():
                circuit_results = []
                for meassurements, count in sub_result.items():
                    # separate end measurements from mid-circuit measurements
                    if meassurements == " ":
                        separate_measurements = [meassurements.split(" ")[0]]
                    else:
                        separate_measurements = meassurements.split(" ")

                    # map to eigenvalues
                    result_eigenvalues = [
                        np.array([-1 if x == "0" else 1 for x in i])
                        for i in separate_measurements
                    ]
                    circuit_results.append(
                        SubResult(result_eigenvalues, count / shots * samples)
                    )
                experiment_run_results.append(circuit_results)
            if group_ind >= len(preocessed_results):
                preocessed_results.append([])
            preocessed_results[group_ind].append(TotalResult(experiment_run_results))
        
    return preocessed_results

def _get_sub_expectation_values(
    experiment_run: TotalResult,
    observables: list,
    shots: int,
    map_qubits: Optional[dict[int, int]] = None,
) -> np.ndarray:
    """Calculate sub expectation value for the result.

    Args:
        experiment_run (TotalResult): results of a subcircuit pair
        observables (list[int | list[int]]):
            list of observables as qubit indices (Z-observables)
        shots (int): number of shots

    Returns:
        list:
            list of sub expectation values

    """
    # generate all possible combinations between end of circuit measurements
    # from subcircuit group
    sub_circuit_result_combinations = product(*experiment_run.subcircuits[0]) # type: ignore[no-matching-overload]

    # initialize sub solution array
    sub_expectation_value = np.zeros(len(observables))

    for ind, circuit_result in enumerate(sub_circuit_result_combinations):  
        # loop through results
        # concat results to one array and reverse to account for qiskit quibit ordering
        full_result = np.concatenate(
            [i.measurements[0] for i in reversed(circuit_result)]
        )

        if full_result.size == 0:
            raise ValueError("No measurement results found. This should not happen.")
            continue
        
        if map_qubits is not None:
            sorted_full_result = np.array(
                [
                    full_result[map_qubits[key]]
                    for key in sorted(map_qubits.keys(), reverse=True)
                ]
            )
        else:
            sorted_full_result = full_result
        
        sorted_full_result = list(reversed(sorted_full_result))

        qpd_measurement_coefficient = 1  # initial value for qpd
        weight = shots  # initial weight
        for res in circuit_result:  # calculate weight and qpd coefficient
            weight *= res.count / shots
            # if len(res.measurements) > 1:
            qpd_measurement_coefficient *= np.prod(res.measurements[1])
        observable_results = np.empty(len(observables))  # initialize empty array
        # for obsrvables
        for count, obs in enumerate(observables):  # populate observable array
            if isinstance(obs, int):
                observable_results[count] = sorted_full_result[obs]  # if single qubit
            # observable just save
            # to array
            else:  # if multi qubit observable
                multi_qubit_observable_eigenvalue = 1  # initial eigenvalue
                for sub_observables in obs:  # multio qubit observable
                    multi_qubit_observable_eigenvalue *= sorted_full_result[
                        sub_observables
                    ]
                    observable_results[count] = (
                        np.power(-1, len(obs) + 1) * multi_qubit_observable_eigenvalue
                    )

        observable_expectation_value = (
            qpd_measurement_coefficient * observable_results * weight
        )
        sub_expectation_value += observable_expectation_value

    return sub_expectation_value

def estimate_expectation_values(
    results: list[list[TotalResult]],
    expv_data: dict
) -> list[float]:
    """Calculate the estimated expectation values.

    Loop through processed results. For each result group generate all products of
    different measurements from different subcircuits of the group. For each result
    from qpd measurements calculate qpd coefficient and from counts calculate weight.
    Get results for qubits corresponding to the observables. If multiqubit observable
    multiply individual qubit eigenvalues and multiply by (-1)^(m+1) where m is number
    of qubits in the observable. Multiply by weight and add to sub expectation value.
    Once all results iterated over move to next circuit group. Lastly multiply
    by total cut cost and divide by number of samples.

    Args:
        results (list[TotalResult]): results from experiment circuits
        coefficients (list[int]): list of coefficients for each subcircuit group
        cut_locations (np.ndarray[CutLocation]): cut locations
        observables (list[int | list[int]]):
            observables to calculate expectation values for

    Returns:
        list[float]:
            expectation values as a list of floats

    """
    cuts = len(expv_data["cut_locations"])
    wire_cuts = len([i for i in expv_data["cut_locations"] 
                      if isinstance(i, SingleQubitCutLocation)])
    cz_cuts = cuts - wire_cuts
    # number of samples neede
    samples = int(
        (np.power(4, 2 * wire_cuts) * np.power(3, 2 * cz_cuts)) / np.power(ERROR, 2)
    )
    shots = int(samples / len(results))

    measurement_settings = _combine_pauli_ops(expv_data["observables"])

    result_for_obs = []

    for obs in expv_data["observables"].paulis:
        obs_circuit_info = _get_observable_circuit_index(obs, measurement_settings)
        result_for_obs.append(obs_circuit_info)

    sum_shots = 0
    # ininialize approx expectation values of an array of ones
    expectation_values = np.ones(len(expv_data["observables"]))

    for ind, obs_data in enumerate(result_for_obs):
        if obs_data["circuit_index"] is None:
            raise ValueError("""Observable cannot be measured 
                             with given measurement settings.""")
        
        for experiment_run, coefficient in zip(results, expv_data["coefficients"]):
        # add sub results to the total approx expectation value
            cur_obs = (obs_data["obs_indices"] 
                       if len(obs_data["obs_indices"]) == 1 
                       else [obs_data["obs_indices"]])
            mid = (
                np.power(-1, wire_cuts + 1)  # * (np.power(-1, cz_cuts)
                * coefficient
                * _get_sub_expectation_values(
                    experiment_run[obs_data["circuit_index"]], cur_obs,
                    shots, expv_data["map_qubit"])
            )[0]
            sum_shots += shots
            expectation_values[ind] += mid

    # multiply by gamma to the power of cuts and take mean
    return np.power(4, wire_cuts) * np.power(3, cz_cuts) * expectation_values / samples


"""
A module for the main circuit knitting workflow.
"""

from __future__ import annotations

import pickle

import numpy as np
from qiskit import QuantumCircuit, transpile
from qiskit.circuit import (
    Qubit,
)
from qiskit.converters import circuit_to_dag
from qiskit.quantum_info import SparsePauliOp
from qiskit_aer import AerSimulator

from QCut.backend_utility import transpile_subcircuits
from QCut.basis_transform import (
    _combine_pauli_ops,
    _get_obs_subcircuits,
)
from QCut.circuit_preparation import get_locations_and_subcircuits
from QCut.circuit_utils import _remove_obsm, _remove_obsm_2
from QCut.cutcircuit import CutCircuit, CutExperiment
from QCut.cutlocation import CutLocation, SingleQubitCutLocation
from QCut.postprocess import ERROR, _process_results, estimate_expectation_values
from QCut.qcutresult import TotalResult
from QCut.qpd_operations import (
    _insert_cz_cut_qpd,
    _insert_wire_cut_qpd,
    get_qpd_combinations,
)


def _finalize_subcircuit(
    subcircuit: QuantumCircuit, qpd_qubits: list[int]
) -> QuantumCircuit:
    """Finalize the subcircuit by measuring remaining qubits and decomposing."""

    meas_qubits = [i for i in range(subcircuit.num_qubits) if i not in qpd_qubits]

    dag = circuit_to_dag(subcircuit)
    idle = list(dag.idle_wires())

    for wire in idle:
        if isinstance(wire, Qubit) and wire._index in meas_qubits:
            meas_qubits.remove(wire._index)

    if len(subcircuit.cregs) >= 2:
        subcircuit.measure(meas_qubits, subcircuit.cregs[1])
    else:
        subcircuit.measure(meas_qubits, subcircuit.cregs[0])
    return subcircuit


def _get_placeholder_locations(subcircuits: list[QuantumCircuit]) -> list:
    """
    Identify the locations of placeholder operations in a list of quantum subcircuits.
    This function scans through each quantum subcircuit provided in the input list and
    identifies the indices and operations where either measurement ("Meas") or
    initialization ("Init") operations occur. It returns a list of lists, where each
    sublist corresponds to  a subcircuit and contains tuples of the
    form (index, operation).

    Args:
        subcircuits (list[QuantumCircuit]):
            A list of QuantumCircuit objects to be analyzed.
    Returns:
        list:
            A list of lists, where each sublist contains tuples (index, operation)
            indicating the positions of measurement or initialization operations in
            the corresponding subcircuit.

    """
    ops = []
    names = ["Meas", "Init", "cutCZ"]
    for circ in subcircuits:
        subops = []
        for ind, op in enumerate(circ.data):
            # if "Meas" in op.operation.name or "Init" in op.operation.name :
            if any(i in op.operation.name for i in names):
                subops.append((ind, op))
        ops.append(subops)

    return ops

def get_experiment_circuits(  # noqa: C901
    cut_circuit: CutCircuit,
    observables: SparsePauliOp,
) -> CutExperiment:
    """Generate experiment circuits by inserting QPD operations on
    measure/initialize/cutCZ nodes.

    Args:
        subcircuits (list[QuantumCircuit]): subcircuits with measure/initialize nodes.
        cut_locations (np.ndarray[CutLocation]): cut locations.

    Returns:
        tuple: A tuple containing:
            - CutCircuit: A CutCircuit object containing the experiment circuits.
            - list[int]: A list of coefficients for each circuit.
            - list[tuple[int, int, int]]:
                A list of index pointers to results that need additional post-processing
                due to identity basis measurement.

    """

    num_qubits = 0
    for subcircuit in cut_circuit.subcircuits:
        crs = subcircuit.cregs
        for cr in crs:
            if cr.name == "meas":
                num_qubits += cr.size

    if all(len(obs) != num_qubits for obs in observables.paulis):
        raise ValueError(
            f"""ALL observable lengths must match 
            the number of qubits in the original uncut circuit 
            ({num_qubits})."""
        )

    qpd_combinations = get_qpd_combinations(cut_circuit.cut_locations)  
    # generate the QPD
    # operation combinations

    check_circuit_type = cut_circuit.backend is not None
    
    measurement_settings = _combine_pauli_ops(observables)

    backend = None
    if check_circuit_type:
        backend = cut_circuit.backend
        try:
            basis = backend.configuration().basis_gates # type: ignore[possibly-missing-attribute]
        except Exception:
            basis = list(backend.architecture.gates.keys()) # type: ignore[possibly-missing-attribute]
        basis = ["r" if gate == "prx" else gate for gate in basis]

    obs_subcircuits = None

    if check_circuit_type:
        x_meas_ops = QuantumCircuit(1)
        x_meas_ops.h(0)
        x_meas_ops.name = "X-meas"
        x_meas_ops = transpile(x_meas_ops, basis_gates=basis)
        x_meas_ops = x_meas_ops.to_instruction()

        y_meas_ops = QuantumCircuit(1)
        y_meas_ops.sdg(0)
        y_meas_ops.h(0)
        y_meas_ops.name = "Y-meas"
        y_meas_ops = transpile(y_meas_ops, basis_gates=basis)
        y_meas_ops = y_meas_ops.to_instruction()

        ops = {"X-meas": x_meas_ops, "Y-meas": y_meas_ops}

        obs_subcircuits = _get_obs_subcircuits(
            cut_circuit.subcircuits, measurement_settings, ops
        )
    else:
        obs_subcircuits = _get_obs_subcircuits(
            cut_circuit.subcircuits, measurement_settings
        )



    _remove_obsm(obs_subcircuits)

    _remove_obsm_2(cut_circuit.subcircuits)

    # initialize solution lists
    cuts = len(cut_circuit.cut_locations)
    cz_cuts = len([i for i in cut_circuit.cut_locations if isinstance(i, CutLocation)])
    wire_cuts = cuts - cz_cuts

    num_circs = np.power(8, wire_cuts) * np.power(6, cz_cuts)
    experiment_circuits = []
    coefficients = np.empty(num_circs)
    placeholder_locations = _get_placeholder_locations(cut_circuit.subcircuits)
    for id_meas_experiment_index, qpd in enumerate(
        qpd_combinations
    ):  # loop through all
        # QPD combinations
        coefficients[id_meas_experiment_index] = np.prod([op["c"] for op in qpd])

        if check_circuit_type:
            for sub in qpd:
                sub["op_0"] = transpile(sub["op_0"], basis_gates=basis)
                sub["op_1"] = transpile(sub["op_1"], basis_gates=basis)

        # circuits
        inserted_operations = 0
        obs_set_circuits = []
        for obs_set in obs_subcircuits:
            cur_set_circuits = {}
            for id_meas_subcircuit_index, circ in obs_set.items():
                subcircuit = pickle.loads(pickle.dumps(circ))
                # subcircuit = deepcopy(circ)
                offset = 0
                classical_bit_index = 0
                qpd_qubits = []  # store the qubit indices of qubits used for qpd
                # measurements
                for op_ind in placeholder_locations[id_meas_subcircuit_index]:
                    ind, op = op_ind

                    actual_op = subcircuit.data[ind + offset]
                    
                    if actual_op.operation.name != op.operation.name:
                        cur_ind = ind + offset
                        for i in range(len(subcircuit.data)):
                            cur_ind_minus = cur_ind - i
                            cur_ind_plus = cur_ind + i
                            if op.operation.name == subcircuit.data[
                                cur_ind_minus].operation.name:
                                ind = cur_ind_minus - offset
                                break
                            if op.operation.name == subcircuit.data[
                                cur_ind_plus].operation.name:
                                ind = cur_ind_plus - offset
                                break
                            

                    if "cut" in op.operation.name:
                        (
                            offset,
                            classical_bit_index,
                            inserted_operations,
                        ) = _insert_cz_cut_qpd(
                            ind,
                            op,
                            subcircuit,
                            offset,
                            qpd_qubits,
                            qpd,
                            classical_bit_index,
                            inserted_operations,
                        )

                    elif "Meas" in op.operation.name or "Init" in op.operation.name:
                        (
                            offset,
                            classical_bit_index,
                            inserted_operations,
                        ) = _insert_wire_cut_qpd(
                            ind,
                            op,
                            subcircuit,
                            offset,
                            qpd_qubits,
                            qpd,
                            classical_bit_index,
                            inserted_operations,
                        )
                    else:
                        raise ValueError(
                            f"""Unknown placeholder operation: {op.operation.name}.
                            Actual operation: {subcircuit.data[ind + offset]}"""
                        )

                subcircuit = _finalize_subcircuit(subcircuit, qpd_qubits)
                cur_set_circuits[id_meas_subcircuit_index] = subcircuit
            obs_set_circuits.append(cur_set_circuits)
        experiment_circuits.append(obs_set_circuits)
    return CutExperiment(
        experiment_circuits,
        cut_circuit.cut_locations,
        cut_circuit.map_qubit,
        coefficients, # type: ignore[invalid-argument-type]
        observables,
        backend=backend,
    )

def run_experiments(  # noqa: C901
    cut_experiment: CutExperiment,
    shots: int = 2**12,
    backend = None,
) -> list[list[TotalResult]]:
    """Run experiment circuits.

    Loop through experiment circuits and then loop through circuit group and run each
    circuit. Store results as [group0, group1, ...] where group is [res0, res1, ...].
    where res is "xxx yy": count xxx are the measurements from the end of circuit
    measurements on the meas classical register and yy are the qpd basis measurement
    results from the qpd_meas class register.

    Args:
        experiment_circuits (CutCircuit): experiment circuits
        cut_locations (np.ndarray[CutLocation]): list of cut locations
        shots (int): number of shots per circuit run (optional)
        backend: backend used for running the circuits (optional)

    Returns:
        list[TotalResult]:
            list of transformed results

    """
    wire_cuts = len([i for i in cut_experiment.cut_locations 
                     if isinstance(i, SingleQubitCutLocation)])
    cz_cuts = len(cut_experiment.cut_locations) - wire_cuts
    samples = int(
        (np.power(4, 2 * wire_cuts) * np.power(3, 2 * cz_cuts)) / np.power(ERROR, 2)
    )
    samples = int(samples / cut_experiment.num_groups)
    if backend is None:
        backend = AerSimulator()

    results = [0] * (cut_experiment.num_groups)

    for count, circuit_group in enumerate(cut_experiment.experiments):
        group = []
        for obs_ind, obs_group in enumerate(circuit_group):
            obs_res = {}
            for ind, subcircuit in obs_group.items():
                try:
                    sub_result = dict(backend.run(
                            subcircuit, shots=shots
                        ).result().get_counts().items())
                except Exception:
                    sub_result = {" " + "0"*subcircuit.num_clbits: shots}
                obs_res[ind] = sub_result
            group.append(obs_res)
        results[count] = group

    all_keys = results[0][0].keys()

    for ind, sub_result in enumerate(results):
        for exp_ind, experiment_run in enumerate(sub_result):
            if experiment_run.keys() != all_keys:
                for key, val in results[0][0].items():
                    if key not in experiment_run:
                        experiment_run[key] = val
    
    return _process_results(results, shots, samples)



def run_cut_circuit(
    cut_circuit: CutCircuit,
    observables: SparsePauliOp,
    backend=AerSimulator(),
) -> list[float]:
    """After splitting the circuit run the rest of the circuit knitting sequence.

    Args:
        subcircuits (list[QuantumCircuit]):
            subcircuits containing the placeholder operations
        cut_locations (np.ndarray[CutLocation]): list of cut locations
        observables (list[int | list[int]]):
            list of observables as qubit indices (Z observable)
        backend: backend to use for running experiment circuits (optional)

    Returns:
        list: a list of expectation values

    """

    if not isinstance(backend, AerSimulator):
        transpiled_subcircuits = transpile_subcircuits(cut_circuit
                                                       ,backend,
                                                       optimization_level=3)
    
        cut_experiment = get_experiment_circuits(transpiled_subcircuits, 
                                        observables)
    else:
        cut_experiment = get_experiment_circuits(cut_circuit, 
                                        observables)
        
    results = run_experiments(
        cut_experiment,
        backend=backend,
    )

    return estimate_expectation_values(
        results, cut_experiment.expv_data()
    )


def run(
    circuit: QuantumCircuit,
    observables: SparsePauliOp,
    backend=AerSimulator(),
) -> list[float]:
    """Run the whole circuit knitting sequence with one function call.

    Args:
        circuit (QuantumCircuit): circuit with cut experiments
        observables (list[int | list[int]]):
            list of observbles in the form of qubit indices (Z-obsevable).
        backend: backend to use for running experiment circuits (optional)

    Returns:
        list: a list of expectation values

    """
    # circuit = circuit.copy()
    cut_circuit = get_locations_and_subcircuits(circuit)

    return run_cut_circuit(cut_circuit, observables, backend)


"""
A module for generating the QPD operations and inserting them into the circuit at 
the appropriate locations.
"""

from __future__ import annotations

from itertools import product
from typing import Iterable

from qiskit.circuit import (
    CircuitInstruction,
    Qubit,
)

from QCut.cutlocation import CutLocation, SingleQubitCutLocation
from QCut.qpd import cz_qpd, identity_qpd


def _insert_wire_cut_qpd(
    ind,
    op,
    subcircuit,
    offset,
    qpd_qubits,
    qpd,
    classical_bit_index,
    inserted_operations,
):
    if "Meas" in op.operation.name:  # if measure channel remove placeholder
        # and insert current
        # qpd operation
        qubit_index = subcircuit.find_bit(op.qubits[0]).index
        subcircuit.data.pop(ind + offset)  # remove plaxceholder
        # measure channel
        qpd_qubits.append(qubit_index)  # store index
        qubits_for_operation = [Qubit(subcircuit.qregs[0], qubit_index)]
        meas_op = qpd[int(op.operation.name.split("_")[-1])]["op_0"]
        if meas_op.name == "id-meas":  # if identity measure channel
            # store indices
            # remove extra classical bits and registers
            #_adjust_cregs(subcircuit)
            for subop in reversed(meas_op.data):
                subcircuit.data.insert(
                    ind + offset,
                    CircuitInstruction(operation=subop.operation, 
                                       qubits=qubits_for_operation),
                )
        else:
            for i, subop in enumerate(reversed(meas_op.data)):
                if i == 0:
                    subcircuit.data.insert(
                        ind + offset,
                        CircuitInstruction(
                            operation=subop.operation,
                            qubits=qubits_for_operation,
                            clbits=[subcircuit.cregs[0][classical_bit_index]],
                        ),
                    )
                else:
                    subcircuit.data.insert(
                        ind + offset,
                        CircuitInstruction(
                            operation=subop.operation,
                            qubits=qubits_for_operation,
                        ),
                    )

            # increment classical bit counter
            classical_bit_index += 1

        inserted_operations += 1
        offset += len(meas_op.data) - 1

    if "Init" in op.operation.name:
        subcircuit.data.pop(ind + offset)
        init_op = qpd[int(op.operation.name.split("_")[-1])]["op_1"]
        qubits_for_operation = [
            Qubit(subcircuit.qregs[0], subcircuit.find_bit(x).index) for x in op.qubits
        ]
        for subop in reversed(init_op.data):
            subcircuit.data.insert(
                ind + offset,
                CircuitInstruction(
                    operation=subop.operation, qubits=qubits_for_operation
                ),
            )

        inserted_operations += 1
        offset += len(init_op) - 1

    return offset, classical_bit_index, inserted_operations


def _insert_cz_cut_qpd(  # noqa: C901
    ind,
    op,
    subcircuit,
    offset,
    qpd_qubits,
    qpd,
    classical_bit_index,
    inserted_operations,
):
    if "_c_" in op.operation.name:  # if measure channel remove placeholder
        # and insert current
        # qpd operation

        qubit_index = subcircuit.find_bit(op.qubits[0]).index
        subcircuit.data.pop(ind + offset)  # remove plaxceholder
        # measure channel
        # qpd_qubits.append(qubit_index)  # store index
        qubits_for_operation = [Qubit(subcircuit.qregs[0], qubit_index)]
        meas_op = qpd[int(op.operation.name.split("_")[-1])]["op_0"]
        if meas_op.name in ["id-meas", "s", "sdg", "z"]:
            # if identity measure channel
            # store indices
            # remove extra classical bits and registers
            #if meas_op.name != "id-meas":
            #    _adjust_cregs(subcircuit)
            for subop in reversed(meas_op.data):
                subcircuit.data.insert(
                    ind + offset,
                    CircuitInstruction(
                        operation=subop.operation, qubits=qubits_for_operation
                    ),
                )
        else:
            for i, subop in enumerate(reversed(meas_op.data)):
                if i == 0:
                    subcircuit.data.insert(
                        ind + offset,
                        CircuitInstruction(
                            operation=subop.operation,
                            qubits=qubits_for_operation,
                            clbits=[subcircuit.cregs[0][classical_bit_index]],
                        ),
                    )
                else:
                    subcircuit.data.insert(
                        ind + offset,
                        CircuitInstruction(
                            operation=subop.operation,
                            qubits=qubits_for_operation,
                        ),
                    )

            # increment classical bit counter
            classical_bit_index += 1

        inserted_operations += 1
        offset += len(meas_op) - 1

    if "_t_" in op.operation.name:
        # and insert current
        # qpd operation
        qubit_index = subcircuit.find_bit(op.qubits[0]).index
        subcircuit.data.pop(ind + offset)  # remove plaxceholder
        # measure channel
        # qpd_qubits.append(qubit_index)  # store index
        qubits_for_operation = [Qubit(subcircuit.qregs[0], qubit_index)]
        meas_op = qpd[int(op.operation.name.split("_")[-1])]["op_1"]
        if meas_op.name in ["id-meas", "s", "sdg", "z"]:
            # if identity measure channel
            # store indices
            
            # remove extra classical bits and registers
            #if meas_op.name != "id-meas":
            #    _adjust_cregs(subcircuit)
            for subop in reversed(meas_op.data):
                subcircuit.data.insert(
                    ind + offset,
                    CircuitInstruction(
                        operation=subop.operation, qubits=qubits_for_operation
                    ),
                )
        else:
            for i, subop in enumerate(reversed(meas_op.data)):
                if i == 0:
                    subcircuit.data.insert(
                        ind + offset,
                        CircuitInstruction(
                            operation=subop.operation,
                            qubits=qubits_for_operation,
                            clbits=[subcircuit.cregs[0][classical_bit_index]],
                        ),
                    )
                else:
                    subcircuit.data.insert(
                        ind + offset,
                        CircuitInstruction(
                            operation=subop.operation,
                            qubits=qubits_for_operation,
                        ),
                    )

            # increment classical bit counter
            classical_bit_index += 1

        inserted_operations += 1
        offset += len(meas_op) - 1

    return offset, classical_bit_index, inserted_operations

def get_qpd_combinations(
    cut_locations: list[CutLocation | SingleQubitCutLocation],
) -> Iterable[tuple[dict]]:
    """Get all possible combinations of the QPD operations so that each combination
    has len(cut_locations) elements.

    Args:
        cut_locations (list[CutLocation | SingleQubitCutLocation]): cut locations

    Returns:
        Iterable[tuple[dict]]:
            Iterable of the possible QPD operations

    """
    qpd_lists = []
    for cut in cut_locations:
        if isinstance(cut, SingleQubitCutLocation):
            qpd_lists.append(identity_qpd)
        elif isinstance(cut, CutLocation):
            qpd_lists.append(cz_qpd)
        else:
            raise TypeError(f"Unknown cut type: {type(cut)}")

    # Cartesian product across all qpd options per cut
    all_combinations = product(*qpd_lists)
    return all_combinations



"""
QCutFind is a QCut module for automatically finding cut locations in a quantum circuit. 
It includes functions for extracting cut data from a graph representation of the
circuit, inserting or appending instructions to a circuit, and refining cut locations.

QCutFind uses the Metis graph partitioning library to find initial cut locations,
and then refines those cuts using a custom refinement algorithm. The resulting cut
locations can be used to partition the circuit into subcircuits for execution on
quantum hardware or simulators.
"""

from QCut.QCutFind.combine_subcircuits import construct_final_subcircuits
from QCut.QCutFind.cut_finding import find_cuts

__all__ = ["find_cuts", "construct_final_subcircuits"]



"Define QCutError class for custom error handling."

from typing import Optional


class QCutError(Exception):
    """Exception raised for custom error conditions.

    Attributes:
        message (str): Explanation of the error.
        code (int, optional): Error code representing the error type.

    """

    def __init__(
        self, message: str = "An error occurred", code: Optional[int] = None
    ) -> None:
        """Init.

        Args:
            message (str): Explanation of the error. Default is "An error occurred".
            code (int): Optional error code representing the error type.

        """
        self.message = message
        self.code = code
        super().__init__(self.message)

    def __str__(self) -> str:
        """Return the string representation of the error.

        Returns
            str: A string describing the error, including the code if available.

        """
        if self.code:
            return f"[Error {self.code}] {self.message}"

        return self.message


"""
Utility functions for converting quantum circuits to graph representations and
vice versa, as well as functions for updating node and edge information in the graph.
"""

import rustworkx as rx


def weight_fn(edge_data):
    """
    Custom weight function for edges.
    """
    if len(edge_data) > 2:
        # your existing logic: tuple[3] is raw_weight
        return edge_data[3]  # Use the weight from the tuple
    # default weight
    return edge_data[1]


max_added = None


def update_added(added_set, added, op):
    """
    Append the two new node indices for a 2-qubit op,
    using a global next-available counter instead of
    scanning `added`.
    """
    global max_added

    q0, q1 = op[1][2]  # the two qubit indices

    # lazy initialize on first call
    if max_added is None:
        if added:
            max_added = max(added)
        else:
            # for the very first gate, we'll set it below
            max_added = None

    # both qubits have been seen before → allocate two brand-new nodes
    if q0 in added_set and q1 in added_set:
        a = max_added + 1 # type: ignore[unsupported-operator]
        b = max_added + 2 # type: ignore[unsupported-operator]
        added.extend([a, b])
        added_set.add(a)
        added_set.add(b)
        max_added += 2 # type: ignore[unsupported-operator]

    # only q0 seen → reuse q1 and allocate one fresh node
    elif q0 in added_set:
        a = q1
        max_added = max(max_added, a) # type: ignore[invalid-argument-type]
        b = max_added + 1
        added.extend([a, b])
        added_set.add(b)
        added_set.add(a)
        max_added += 1

    # only q1 seen → reuse q0 and allocate one fresh node
    elif q1 in added_set:
        a = q0
        max_added = max(max_added, a)  # type: ignore[invalid-argument-type]
        b = max_added + 1
        added.extend([a, b])
        added_set.add(b)
        added_set.add(a)
        max_added += 1

    # neither seen → just append those original qubit-IDs
    else:
        added.extend([q0, q1])
        added_set.add(q0)
        added_set.add(q1)
        # now that they’re in, prime max_added for future
        max_added = max(q0, q1)


def update_nodes_on_qubit(nodes_on_qubit, qubit, node):
    if qubit not in nodes_on_qubit:
        nodes_on_qubit[qubit] = [node]
    else:
        nodes_on_qubit[qubit].append(node)


def circ_to_graph(circuit, gateCutWeight=3, wireCutWeight=4):  # noqa: C901
    """
    Convert a quantum circuit to a graph representation.

    Args:
        circuit: A quantum circuit object.

    Returns:
        A graph representation of the circuit.
    """

    nodes_on_qubit = {}

    ops_2qb = [
        (
            idx,
            (
                x.operation.name,
                x.operation.num_qubits,
                [circuit.find_bit(i).index for i in x.qubits],
            ),
        )
        for idx, x in enumerate(circuit.data)
        if x.operation.num_qubits == 2
    ]

    G = rx.PyGraph()
    indxs = list(
        range(
            max(
                max([x[0] for x in ops_2qb]),
                len(ops_2qb) * 2,
            )
            + 1
        )
    )
    for i in indxs:
        G.add_node(i)

    track_nodes = {}
    added = []
    added_set = set()

    for ind, op in enumerate(ops_2qb):
        if (
            op[1][2][0] not in track_nodes.keys()
            and op[1][2][1] not in track_nodes.keys()
        ):
            update_added(added_set, added, op)
            if op[1][2][0] in added_set:
                track_nodes[op[1][2][0]] = [(added[-2])]
            else:
                track_nodes[op[1][2][0]] = []
            if op[1][2][1] in added_set:
                track_nodes[op[1][2][1]] = [added[-1]]
            else:
                track_nodes[op[1][2][1]] = []

            update_nodes_on_qubit(nodes_on_qubit, op[1][2][0], added[-2])
            update_nodes_on_qubit(nodes_on_qubit, op[1][2][1], added[-1])

            G.add_edge(
                added[-2],
                added[-1],
                (op[0], op[1][0], op[1][2], gateCutWeight),
            )
        elif op[1][2][1] not in track_nodes.keys():
            update_added(added_set, added, op)

            if op[1][2][1] in added_set:
                track_nodes[op[1][2][1]] = [(added[-1])]
            else:
                track_nodes[op[1][2][1]] = []

            track_nodes[op[1][2][0]].append(added[-2])
            ind = track_nodes[op[1][2][0]]

            update_nodes_on_qubit(nodes_on_qubit, op[1][2][0], added[-2])
            update_nodes_on_qubit(nodes_on_qubit, op[1][2][1], added[-1])

            G.add_edge(
                added[-2],
                added[-1],
                (op[0], op[1][0], op[1][2], gateCutWeight),
            )
            if ind == []:
                G.add_edge(op[1][2][0], added[-1], (op[1][2][0], wireCutWeight))
            else:
                if len(ind) == 1:
                    G.add_edge(ind[-1], op[1][2][0], (op[1][2][0], wireCutWeight))
                else:
                    G.add_edge(ind[-1], ind[-2], (op[1][2][0], wireCutWeight))
        elif op[1][2][0] not in track_nodes.keys():
            update_added(added_set, added, op)

            if op[1][2][0] in added_set:
                track_nodes[op[1][2][0]] = [(added[-2])]
            else:
                track_nodes[op[1][2][0]] = []

            track_nodes[op[1][2][1]].append(added[-1])
            ind = track_nodes[op[1][2][1]]

            update_nodes_on_qubit(nodes_on_qubit, op[1][2][0], added[-2])
            update_nodes_on_qubit(nodes_on_qubit, op[1][2][1], added[-1])

            G.add_edge(
                added[-2],
                added[-1],
                (op[0], op[1][0], op[1][2], gateCutWeight),
            )
            if ind == []:
                G.add_edge(op[1][2][1], added[-1], (op[1][2][1], wireCutWeight))
            else:
                if len(ind) == 1:
                    G.add_edge(ind[-1], op[1][2][1], (op[1][2][1], wireCutWeight))
                else:
                    G.add_edge(ind[-1], ind[-2], (op[1][2][1], wireCutWeight))
        else:
            update_added(added_set, added, op)
            track_nodes[op[1][2][0]].append(added[-2])
            track_nodes[op[1][2][1]].append(added[-1])
            ind0 = track_nodes[op[1][2][0]]
            ind1 = track_nodes[op[1][2][1]]

            update_nodes_on_qubit(nodes_on_qubit, op[1][2][0], added[-2])
            update_nodes_on_qubit(nodes_on_qubit, op[1][2][1], added[-1])

            G.add_edge(
                added[-2],
                added[-1],
                (op[0], op[1][0], op[1][2], gateCutWeight),
            )
            if ind0 == []:
                G.add_edge(op[1][2][0], added[-1], (op[1][2][0], wireCutWeight))
            else:
                if len(ind0) == 1:
                    G.add_edge(ind0[-1], op[1][2][0], (op[1][2][0], wireCutWeight))
                else:
                    G.add_edge(ind0[-1], ind0[-2], (op[1][2][0], wireCutWeight))
            if ind1 == []:
                G.add_edge(op[1][2][1], added[-2], (op[1][2][1], wireCutWeight))
            else:
                if len(ind1) == 1:
                    G.add_edge(ind1[-1], op[1][2][1], (op[1][2][1], wireCutWeight))
                else:
                    G.add_edge(ind1[-1], ind1[-2], (op[1][2][1], wireCutWeight))

    # Remove nodes with degree 0 (not connected to anything)
    isolated_nodes = [node for node in G.node_indices() if G.degree(node) == 0]

    G.remove_nodes_from(isolated_nodes)
    G, nodes_on_qubit = final_graph(G, nodes_on_qubit)
    return G, nodes_on_qubit


def final_graph(graph, nodes_on_qubit):
    # Relabel nodes to ensure consecutive numbering
    # Since rustworkx doesn't have relabel_nodes, we need to create a new graph manually
    old_to_new = {
        old_idx: new_idx for new_idx, old_idx in enumerate(graph.node_indices())
    }

    # Create a new graph
    new_graph = rx.PyGraph()

    # Add nodes with consecutive indices (0, 1, 2, ...)
    for i in range(len(graph.node_indices())):
        new_graph.add_node(i)

    # Add edges with remapped indices
    for edge in graph.edge_list():
        old_source, old_target = edge
        new_source = old_to_new[old_source]
        new_target = old_to_new[old_target]
        edge_data = graph.get_edge_data(old_source, old_target)
        new_graph.add_edge(new_source, new_target, edge_data)

    # Update nodes_on_qubit dictionary with new node indices
    updated_nodes_on_qubit = {}
    for qubit, old_node_list in nodes_on_qubit.items():
        updated_nodes_on_qubit[qubit] = [
            old_to_new[old_node] for old_node in old_node_list if old_node in old_to_new
        ]

    return new_graph, updated_nodes_on_qubit



from qiskit import QuantumCircuit
from qiskit.circuit import Parameter
from qiskit.circuit.library import CXGate
from qiskit.quantum_info import SparsePauliOp

import QCut as ck
from QCut import cut, cutGate

cut_qc = QuantumCircuit(4)

mult = Parameter('mult')
cut_qc.r(mult*0.46262, mult*0.1446, 0)
cut_qc.append(**cutGate(CXGate(), 0, 1)) 
cut_qc.append(cut(), [1])
cut_qc.cx(1,2)
cut_qc.cx(2,3)


cut_circuit = ck.get_locations_and_subcircuits(cut_qc)
num_subcircuits = 3

num_qubits = [1,1,3]

def test_cut_circuit_properties():
    assert cut_circuit.num_qubits == num_qubits
    assert cut_circuit.num_subcircuits == num_subcircuits

def test_cut_circuit_assign_parameters():
    cut_circuit_param = cut_circuit.assign_parameters({'mult': 2}, inplace=False)

    for exp_ind, subcircuits in enumerate(cut_circuit_param.subcircuits):
        for circ_ind, value in enumerate(subcircuits):
            for par in value.params:
                if hasattr(par, 'parameters'):
                    for elem in par.parameters:
                        assert 'mult' not in elem.name
                else:
                    assert True

observables = SparsePauliOp(["IIIZ", "IIZI", "IZII", "IIZZ"])
cut_experiment = ck.get_experiment_circuits(cut_circuit, observables)

exp_num_qubits = [1,1,3]
exp_num_circuits = 144
exp_group_size = 3
exp_num_groups = 48
exp_num_obs_groups = 1

def test_cut_experiment_properties():
    assert cut_experiment.num_qubits == exp_num_qubits
    assert cut_experiment.num_circuits == exp_num_circuits
    assert cut_experiment.group_size == exp_group_size
    assert cut_experiment.num_groups == exp_num_groups
    assert cut_experiment.num_obs_groups == exp_num_obs_groups

def test_cut_experiment_assign_parameters():
    cut_experiment_param = cut_experiment.assign_parameters({'mult': 2}, inplace=False)

    for exp_ind, subcircuits in enumerate(cut_experiment_param.experiments):
        for obs_circ in subcircuits:
            for circ in obs_circ.values():
                for circ_ind, value in enumerate(circ):
                    for par in value.params:
                        if hasattr(par, 'parameters'):
                            for elem in par.parameters:
                                assert 'mult' not in elem.name
                        else:
                            assert True

   



"""
Utility functions for working with quantum circuits in the context of circuit 
cutting and knitting.
"""

from qiskit import QuantumCircuit
from qiskit.circuit import Qubit


def _count_gates(circuit: QuantumCircuit) -> dict[Qubit, int]:
    """Count the number of gates acting on each qubit in a QuantumCircuit.

    Args:
        circuit (QuantumCircuit): The input quantum circuit.

    Returns:
        dict[Qubit, int]: A dictionary mapping each qubit to the number of gates 
        acting on it.
    """
    gate_count = dict.fromkeys(circuit.qubits, 0)
    for instruction in circuit.data:
        for qubit in instruction.qubits:
            gate_count[qubit] += 1

    return gate_count


def _remove_idle_wires(circuit: QuantumCircuit) -> QuantumCircuit:
    """Remove idle wires from a QuantumCircuit.

    Args:
        circuit (QuantumCircuit): The input quantum circuit.

    Returns:
        QuantumCircuit: A new quantum circuit with idle wires removed.
    """
    gate_count = _count_gates(circuit)
    for qubit, count in gate_count.items():
        if count == 0:
            circuit.qubits.remove(qubit)
    
    return circuit

def _remove_obsm(subcircuits: list[dict[int, QuantumCircuit]]
                 ):

    for obs_set in subcircuits:
        for ind, circ in obs_set.items():
            j = 0
            while j < len(circ.data):
                if "obs" in circ[j].operation.name:
                    circ.data.remove(circ[j])
                else:
                    j += 1

def _remove_obsm_2(subcircuits: list[QuantumCircuit]
                 ):

    for circ in subcircuits:
        j = 0
        while j < len(circ.data):
            if "obs" in circ[j].operation.name:
                circ.data.remove(circ[j])
            else:
                j += 1


"""Tests for CircuitKnitting package."""  # noqa: N999

from qiskit import QuantumCircuit
from qiskit_aer import AerSimulator

#import QCut as ck
import tests.solutions_automatic_cuts as sq
from QCut import find_cuts
from QCut.circuit_knitting import run_cut_circuit


def test_find_cuts() -> None:
    """Test find_cuts function.

    This function tests whether the find_cuts method correctly identifies the cut
    locations in the provided test circuits by comparing the result to the
    pre-defined solutions.
    """
    for solution_index, circ in enumerate(sq.test_circuits):

        cut_circuit = find_cuts(circ.copy(), 
                                sq.cut_sizes[solution_index], 
                                cuts="both")

        assert len(cut_circuit.subcircuits) == sq.cut_sizes[solution_index]

circuit  =  QuantumCircuit(4)

mult = 1.635
circuit.r(mult*0.46262, mult*0.1446, 0)
circuit.cx(0,1)
circuit.cx(1,2)
circuit.cx(2,3)

def test_find_gate_cuts():
    """Test find_cuts function on a circuit with a cut gate.

    This function tests whether the find_cuts method correctly identifies the cut
    locations in a circuit containing a cut gate by comparing the result to the
    expected number of subcircuits.
    """
    cut_circuit = find_cuts(circuit.copy(), 2, cuts="both")

    assert len(cut_circuit.subcircuits) == 2

    for circ in cut_circuit.subcircuits:
        for op in circ.data:
            assert "meas" not in op.operation.name.lower()
            assert "init" not in op.operation.name.lower()

def test_expectation_values() -> None:
    """Test the expectation values of the test circuits.

    This function tests whether the run method correctly calculates the expectation
    values for each test circuit and its corresponding observable by comparing the
    results to the pre-defined solutions within a specified error tolerance.

    The test runs each circuit on the AerSimulator backend without error mitigation.
    """
    # Initialize the simulator
    sim = AerSimulator()

    # Iterate over each test circuit and its corresponding expected solutions
    for solution_index, circ in enumerate(sq.test_circuits):
        print(solution_index)

        cut_circuit = find_cuts(circ.copy(), 
                                                          sq.cut_sizes[solution_index], 
                                                          cuts="both")

        # Calculate expectation values using the run method
        estimated_expectation_values = run_cut_circuit(cut_circuit, 
                                                          sq.test_observables[solution_index], 
                                                          sim)
        # Check each calculated expectation value against the corresponding
        # expected value
        tolerance = 0.1
        print(estimated_expectation_values)
        print(sq.exp_val_solutions[solution_index])
        for check in [
            abs(a - b) <= tolerance
            for a, b in zip(estimated_expectation_values, 
                            sq.exp_val_solutions[solution_index])
        ]:
            assert check  # noqa: S101
